#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/time.h>

/**
 * Constants
 */

#define BLOCK_SIZE		(8192)

/**
 * Prototypes
 */

	// Creates the test file filled with random data
	void createTestFile(char *testFileName, long long fileSize);

	// Handles runtime errors by displaying the function, activity and error number
	void handleError(char *functionName, char *activity);

	// Standard entry point
	int main(int argc, char *args[]);

	// Prints correct usage and quits
	void printUsageAndQuit();

	// Tests performance of random reads of the given file
	void testRandom(char *testFileName, long long amountToRead);

	// Tests performance of sequential reads of the given file
	void testSeq(char *testFileName);

/**
 * Definitions
 */

/**
 * createTestFile()
 */
void createTestFile(char *testFileName, long long fileSize)
{
	FILE *testFile;
	long long reps, i, j, bufferReps;
	time_t timetmp;
	long long *buffer;
	size_t written;

	// Indicate op
	printf("Creating test file %s of %lld mb\n",testFileName,fileSize);

	// Adjust file size to bytes
	fileSize *= (1024*1024);

	// Allocate a buffer for writing out random long longs
	if (!(buffer = malloc(BLOCK_SIZE)))
		handleError("createTestFile()","malloc");

	// Open the file for writing
	if (!(testFile = fopen(testFileName, "wb")))
		handleError("createTestFile()","fopen");

	// Initialise the random number generator
	srandom(time(NULL));

	// Write data
	reps 		= fileSize / BLOCK_SIZE;
	bufferReps	= BLOCK_SIZE / sizeof(long long);
	for (i = 0; i < reps; i++)
	{
		// Fill buffer with random data
		for (j = 0; j < bufferReps; j++)
			buffer[j] = random();

		// Write
		written = fwrite(buffer, sizeof(long long), bufferReps, testFile);
		if (written != bufferReps)
			handleError("createTestFile()","fwrite");
	}

	// Flush and close
	if (fflush(testFile))
		handleError("createTestFile()","fflush");
	if (fclose(testFile))
		handleError("createTestFile()","fclose");

	// Free buffer
	free(buffer);
}

/**
 * handleError()
 */
void handleError(char *functionName, char *activity)
{
	fprintf(stderr, "Error in %s while attempting %s. Error %d (%s)\n", functionName, activity, errno, strerror(errno));
	exit(1);
}

/**
 * main()
 */
int main(int argc, char *argv[])
{
	// Print usage and quit if argument count is definitely incorrect
	if (argc < 3)
	{
		// Definitely wrong
		printUsageAndQuit();
	}
	else
	{
		// Dispatch
		if (!strcmp(argv[1], "create"))
		{
			if (argc != 4)
				printUsageAndQuit();

			// Create the test file of the specified size
			createTestFile(argv[2], atol(argv[3]));
		}
		else if (!strcmp(argv[1], "seqtest"))
		{
			if (argc != 3)
				printUsageAndQuit();

			// Test performance of sequential reads
			testSeq(argv[2]);
		}
		else if (!strcmp(argv[1], "rndtest"))
		{
			if (argc != 4)
				printUsageAndQuit();

			// Test performance of random reads
			testRandom(argv[2], atol(argv[3]));
		}
		else
		{
			// Unknown command
			printUsageAndQuit();
		}
	}

	return 0;
}

/**
 * printUsageAndQuit()
 */
void printUsageAndQuit()
{
	puts("USAGE: rndpgcst [create <file> <size_in_mb>] | [seqtest <file>] | [rndtest <file> <read_in_mb>]");

	exit(1);
}

/**
 * testSeq()
 */
void testSeq(char *testFileName)
{
	FILE *testFile;
	char *buffer;
	long long reps, totalRead, thisRead, timeTaken;
	struct timeval startTime, endTime;
	struct timezone timezoneDiscard;

	// Indicate op
	printf("Sequential read test of %s\n",testFileName);

	// Grab a buffer
	buffer = malloc(BLOCK_SIZE);

	// Open the file for reading
	if (!(testFile = fopen(testFileName, "rb")))
		handleError("testSeq()","fopen");

	// Start timer
	if (gettimeofday(&startTime, &timezoneDiscard) == -1)
		handleError("testSeq()", "gettimeofday start");

	// Read all data from file
	totalRead = 0;
	while ((thisRead = fread(buffer, 1, BLOCK_SIZE, testFile)) != 0)
		totalRead += thisRead;

	// End timer
	if (gettimeofday(&endTime, &timezoneDiscard) == -1)
		handleError("testSeq()", "gettimeofday start");

	// Close
	if (fclose(testFile))
		handleError("testSeq()","fclose");

	// Free the buffer
	free(buffer);

	// Display time taken
	timeTaken = (endTime.tv_sec - startTime.tv_sec) * 1000000;
	timeTaken += (endTime.tv_usec - startTime.tv_usec);
	printf("%lld bytes read in %f seconds\n", totalRead, (double) timeTaken / (double) 1000000);
}

/**
 * testRandom()
 */
void testRandom(char *testFileName, long long amountToRead)
{
	FILE *testFile;
	long long reps, i, fileSize, timeTaken, totalRead, readPos, thisRead, offsetMax;
	struct stat fileStats;
	char *buffer;
	struct timeval startTime, endTime;
	struct timezone timezoneDiscard;

	// Indicate op
	printf("Random read test of %s for %lld mb\n", testFileName, amountToRead);

	// Initialise the random number generator
	srandom(time(NULL));

	// Adjust amount to read
	amountToRead *= (1024*1024);

	// Determine file size
	if (stat(testFileName, &fileStats) == -1)
		handleError("testRandom()", "stat");
	fileSize = fileStats.st_size;

	// Grab a buffer
	buffer = malloc(BLOCK_SIZE);

	// Open the file for reading
	if (!(testFile = fopen(testFileName, "rb")))
		handleError("testRandom()","fopen");

	// Start timer
	if (gettimeofday(&startTime, &timezoneDiscard) == -1)
		handleError("testRandom()", "gettimeofday start");

	// Read data from file
	reps 		= amountToRead / BLOCK_SIZE;
	offsetMax	= fileSize / BLOCK_SIZE;
	for (i = 0; i < reps; i++)
	{
		// Determine read position
		readPos = (random() % offsetMax) * BLOCK_SIZE;

		// Seek and read
		if (fseek(testFile, readPos, SEEK_SET) == -1)
			handleError("testRandom()","fseek");
		if ((thisRead = fread(buffer, 1, BLOCK_SIZE, testFile)) != BLOCK_SIZE)
			handleError("testRandom()","fread");
	}

	// End timer
	if (gettimeofday(&endTime, &timezoneDiscard) == -1)
		handleError("testRandom()", "gettimeofday start");

	// Close
	if (fclose(testFile))
		handleError("testRandom()","fclose");

	// Free the buffer
	free(buffer);

	// Display time taken
	timeTaken = (endTime.tv_sec - startTime.tv_sec) * 1000000;
	timeTaken += (endTime.tv_usec - startTime.tv_usec);
	printf("%lld bytes read in %f seconds\n", amountToRead, (double) timeTaken / (double) 1000000);
}
